# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Form widgets

This module defines some custom widgets for use with WuttaWeb.

However for convenience it also makes other Deform widgets available
in the namespace:

* :class:`deform:deform.widget.Widget` (base class)
* :class:`deform:deform.widget.TextInputWidget`
* :class:`deform:deform.widget.TextAreaWidget`
* :class:`deform:deform.widget.PasswordWidget`
* :class:`deform:deform.widget.CheckedPasswordWidget`
* :class:`deform:deform.widget.CheckboxWidget`
* :class:`deform:deform.widget.SelectWidget`
* :class:`deform:deform.widget.CheckboxChoiceWidget`
* :class:`deform:deform.widget.DateTimeInputWidget`
* :class:`deform:deform.widget.MoneyInputWidget`
"""

import datetime
import os

import colander
import humanize
from deform.widget import (Widget, TextInputWidget, TextAreaWidget,
                           PasswordWidget, CheckedPasswordWidget,
                           CheckboxWidget, SelectWidget, CheckboxChoiceWidget,
                           DateTimeInputWidget, MoneyInputWidget)
from webhelpers2.html import HTML

from wuttjamaican.conf import parse_list

from wuttaweb.db import Session
from wuttaweb.grids import Grid


class ObjectRefWidget(SelectWidget):
    """
    Widget for use with model "object reference" fields, e.g.  foreign
    key UUID => TargetModel instance.

    While you may create instances of this widget directly, it
    normally happens automatically when schema nodes of the
    :class:`~wuttaweb.forms.schema.ObjectRef` (sub)type are part of
    the form schema; via
    :meth:`~wuttaweb.forms.schema.ObjectRef.widget_maker()`.

    In readonly mode, this renders a ``<span>`` tag around the
    :attr:`model_instance` (converted to string).

    Otherwise it renders a select (dropdown) element allowing user to
    choose from available records.

    This is a subclass of :class:`deform:deform.widget.SelectWidget`
    and uses these Deform templates:

    * ``select``
    * ``readonly/objectref``

    .. attribute:: model_instance

       Reference to the model record instance, i.e. the "far side" of
       the foreign key relationship.

       .. note::

          You do not need to provide the ``model_instance`` when
          constructing the widget.  Rather, it is set automatically
          when the :class:`~wuttaweb.forms.schema.ObjectRef` type
          instance (associated with the node) is serialized.
    """
    readonly_template = 'readonly/objectref'

    def __init__(self, request, url=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.url = url

    def get_template_values(self, field, cstruct, kw):
        """ """
        values = super().get_template_values(field, cstruct, kw)

        # add url, only if rendering readonly
        readonly = kw.get('readonly', self.readonly)
        if readonly:
            if 'url' not in values and self.url and getattr(field.schema, 'model_instance', None):
                values['url'] = self.url(field.schema.model_instance)

        return values


class NotesWidget(TextAreaWidget):
    """
    Widget for use with "notes" fields.

    In readonly mode, this shows the notes with a background to make
    them stand out a bit more.

    Otherwise it effectively shows a ``<textarea>`` input element.

    This is a subclass of :class:`deform:deform.widget.TextAreaWidget`
    and uses these Deform templates:

    * ``textarea``
    * ``readonly/notes``
    """
    readonly_template = 'readonly/notes'


class WuttaCheckboxChoiceWidget(CheckboxChoiceWidget):
    """
    Custom widget for :class:`python:set` fields.

    This is a subclass of
    :class:`deform:deform.widget.CheckboxChoiceWidget`, but adds
    Wutta-related params to the constructor.

    :param request: Current :term:`request` object.

    :param session: Optional :term:`db session` to use instead of
       :class:`wuttaweb.db.sess.Session`.

    It uses these Deform templates:

    * ``checkbox_choice``
    * ``readonly/checkbox_choice``
    """

    def __init__(self, request, session=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()
        self.session = session or Session()


class WuttaDateTimeWidget(DateTimeInputWidget):
    """
    Custom widget for :class:`python:datetime.datetime` fields.

    The main purpose of this widget is to leverage
    :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.render_datetime()`
    for the readonly display.

    It is automatically used for SQLAlchemy mapped classes where the
    field maps to a :class:`sqlalchemy:sqlalchemy.types.DateTime`
    column.  For other (non-mapped) datetime fields, you may have to
    use it explicitly via
    :meth:`~wuttaweb.forms.base.Form.set_widget()`.

    This is a subclass of
    :class:`deform:deform.widget.DateTimeInputWidget` and uses these
    Deform templates:

    * ``datetimeinput``
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get('readonly', self.readonly)
        if readonly and cstruct:
            dt = datetime.datetime.fromisoformat(cstruct)
            return self.app.render_datetime(dt)

        return super().serialize(field, cstruct, **kw)


class FileDownloadWidget(Widget):
    """
    Widget for use with :class:`~wuttaweb.forms.schema.FileDownload`
    fields.

    This only supports readonly, and shows a hyperlink to download the
    file.  Link text is the filename plus file size.

    This is a subclass of :class:`deform:deform.widget.Widget` and
    uses these Deform templates:

    * ``readonly/filedownload``

    :param request: Current :term:`request` object.

    :param url: Optional URL for hyperlink.  If not specified, file
       name/size is shown with no hyperlink.
    """
    readonly_template = 'readonly/filedownload'

    def __init__(self, request, *args, **kwargs):
        self.url = kwargs.pop('url', None)
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, field, cstruct, **kw):
        """ """
        # nb. readonly is the only way this rolls
        kw['readonly'] = True
        template = self.readonly_template

        path = cstruct or None
        if path:
            kw.setdefault('filename', os.path.basename(path))
            kw.setdefault('filesize', self.readable_size(path))
            if self.url:
                kw.setdefault('url', self.url)

        else:
            kw.setdefault('filename', None)
            kw.setdefault('filesize', None)

        kw.setdefault('url', None)
        values = self.get_template_values(field, cstruct, kw)
        return field.renderer(template, **values)

    def readable_size(self, path):
        """ """
        try:
            size = os.path.getsize(path)
        except os.error:
            size = 0
        return humanize.naturalsize(size)


class GridWidget(Widget):
    """
    Widget for fields whose data is represented by a :term:`grid`.

    This is a subclass of :class:`deform:deform.widget.Widget` but
    does not use any Deform templates.

    This widget only supports "readonly" mode, is not editable.  It is
    merely a convenience around the grid itself, which does the heavy
    lifting.

    Instead of creating this widget directly you probably should call
    :meth:`~wuttaweb.forms.base.Form.set_grid()` on your form.

    :param request: Current :term:`request` object.

    :param grid: :class:`~wuttaweb.grids.base.Grid` instance, used to
       display the field data.
    """

    def __init__(self, request, grid, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.grid = grid

    def serialize(self, field, cstruct, **kw):
        """
        This widget simply calls
        :meth:`~wuttaweb.grids.base.Grid.render_table_element()` on
        the ``grid`` to serialize.
        """
        readonly = kw.get('readonly', self.readonly)
        if not readonly:
            raise NotImplementedError("edit not allowed for this widget")

        return self.grid.render_table_element()


class RoleRefsWidget(WuttaCheckboxChoiceWidget):
    """
    Widget for use with User
    :attr:`~wuttjamaican:wuttjamaican.db.model.auth.User.roles` field.
    This is the default widget for the
    :class:`~wuttaweb.forms.schema.RoleRefs` type.

    This is a subclass of :class:`WuttaCheckboxChoiceWidget`.
    """
    readonly_template = 'readonly/rolerefs'

    def serialize(self, field, cstruct, **kw):
        """ """
        model = self.app.model

        # special logic when field is editable
        readonly = kw.get('readonly', self.readonly)
        if not readonly:

            # but does not apply if current user is root
            if not self.request.is_root:
                auth = self.app.get_auth_handler()
                admin = auth.get_role_administrator(self.session)

                # prune admin role from values list; it should not be
                # one of the options since current user is not admin
                values = kw.get('values', self.values)
                values = [val for val in values
                          if val[0] != admin.uuid]
                kw['values'] = values

        else: # readonly

            # roles
            roles = []
            if cstruct:
                for uuid in cstruct:
                    role = self.session.get(model.Role, uuid)
                    if role:
                        roles.append(role)
            kw['roles'] = roles

            # url
            url = lambda role: self.request.route_url('roles.view', uuid=role.uuid)
            kw['url'] = url

        # default logic from here
        return super().serialize(field, cstruct, **kw)


class UserRefsWidget(WuttaCheckboxChoiceWidget):
    """
    Widget for use with Role
    :attr:`~wuttjamaican:wuttjamaican.db.model.auth.Role.users` field.
    This is the default widget for the
    :class:`~wuttaweb.forms.schema.UserRefs` type.

    This is a subclass of :class:`WuttaCheckboxChoiceWidget`; however
    it only supports readonly mode and does not use a template.
    Rather, it generates and renders a
    :class:`~wuttaweb.grids.base.Grid` showing the users list.
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get('readonly', self.readonly)
        if not readonly:
            raise NotImplementedError("edit not allowed for this widget")

        model = self.app.model
        columns = ['username', 'active']

        # generate data set for users
        users = []
        if cstruct:
            for uuid in cstruct:
                user = self.session.get(model.User, uuid)
                if user:
                    users.append(dict([(key, getattr(user, key))
                                       for key in columns + ['uuid']]))

        # do not render if no data
        if not users:
            return HTML.tag('span')

        # grid
        grid = Grid(self.request, key='roles.view.users',
                    columns=columns, data=users)

        # view action
        if self.request.has_perm('users.view'):
            url = lambda user, i: self.request.route_url('users.view', uuid=user['uuid'])
            grid.add_action('view', icon='eye', url=url)
            grid.set_link('person')
            grid.set_link('username')

        # edit action
        if self.request.has_perm('users.edit'):
            url = lambda user, i: self.request.route_url('users.edit', uuid=user['uuid'])
            grid.add_action('edit', url=url)

        # render as simple <b-table>
        # nb. must indicate we are a part of this form
        form = getattr(field.parent, 'wutta_form', None)
        return grid.render_table_element(form)


class PermissionsWidget(WuttaCheckboxChoiceWidget):
    """
    Widget for use with Role
    :attr:`~wuttjamaican:wuttjamaican.db.model.auth.Role.permissions`
    field.

    This is a subclass of :class:`WuttaCheckboxChoiceWidget`.  It uses
    these Deform templates:

    * ``permissions``
    * ``readonly/permissions``
    """
    template = 'permissions'
    readonly_template = 'readonly/permissions'

    def serialize(self, field, cstruct, **kw):
        """ """
        kw.setdefault('permissions', self.permissions)

        if 'values' not in kw:
            values = []
            for gkey, group in self.permissions.items():
                for pkey, perm in group['perms'].items():
                    values.append((pkey, perm['label']))
            kw['values'] = values

        return super().serialize(field, cstruct, **kw)


class EmailRecipientsWidget(TextAreaWidget):
    """
    Widget for :term:`email setting` recipient fields (``To``, ``Cc``,
    ``Bcc``).

    This is a subclass of
    :class:`deform:deform.widget.TextAreaWidget`.  It uses these
    Deform templates:

    * ``textarea``
    * ``readonly/email_recips``

    See also the :class:`~wuttaweb.forms.schema.EmailRecipients`
    schema type, which uses this widget.
    """
    readonly_template = 'readonly/email_recips'

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get('readonly', self.readonly)
        if readonly:
            kw['recips'] = parse_list(cstruct or '')

        return super().serialize(field, cstruct, **kw)

    def deserialize(self, field, pstruct):
        """ """
        if pstruct is colander.null:
            return colander.null

        values = [value for value in parse_list(pstruct)
                  if value]
        return ', '.join(values)


class BatchIdWidget(Widget):
    """
    Widget for use with the
    :attr:`~wuttjamaican:wuttjamaican.db.model.batch.BatchMixin.id`
    field of a :term:`batch` model.

    This widget is "always" read-only and renders the Batch ID as
    zero-padded 8-char string
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        if cstruct is colander.null:
            return colander.null

        batch_id = int(cstruct)
        return f'{batch_id:08d}'



# -*- coding: utf-8; -*-

import datetime
from unittest import TestCase
from unittest.mock import patch

import sqlalchemy as sa

from wuttjamaican.db.model import Base

from wuttaweb.grids import filters as mod
from tests.util import WebTestCase


class TestGridFilter(WebTestCase):

    def setUp(self):
        self.setup_web()

        model = self.app.model
        self.sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'ggg'},
            {'name': 'foo4', 'value': 'ggg'},
            {'name': 'foo5', 'value': 'ggg'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in self.sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        self.sample_query = self.session.query(model.Setting)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop('factory', mod.GridFilter)
        kwargs['model_property'] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_constructor(self):
        model = self.app.model

        # verbs is not set by default, but can be set
        filtr = self.make_filter(model.Setting.name)
        self.assertFalse(hasattr(filtr, 'verbs'))
        filtr = self.make_filter(model.Setting.name, verbs=['foo', 'bar'])
        self.assertEqual(filtr.verbs, ['foo', 'bar'])

        # verb is not set by default, but can be set
        filtr = self.make_filter(model.Setting.name)
        self.assertFalse(hasattr(filtr, 'verb'))
        filtr = self.make_filter(model.Setting.name, verb='foo')
        self.assertEqual(filtr.verb, 'foo')

        # default verb is not set by default, but can be set
        filtr = self.make_filter(model.Setting.name)
        self.assertFalse(hasattr(filtr, 'default_verb'))
        filtr = self.make_filter(model.Setting.name, default_verb='foo')
        self.assertEqual(filtr.default_verb, 'foo')

    def test_repr(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.GridFilter)
        self.assertEqual(repr(filtr), "GridFilter(key='name', active=False, verb=None, value=None)")

    def test_get_verbs(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)
        self.assertFalse(hasattr(filtr, 'verbs'))
        self.assertEqual(filtr.default_verbs, ['equal', 'not_equal'])

        # by default, returns default verbs (plus 'is_any')
        self.assertEqual(filtr.get_verbs(), ['equal', 'not_equal', 'is_any'])

        # default verbs can be a callable
        filtr.default_verbs = lambda: ['foo', 'bar']
        self.assertEqual(filtr.get_verbs(), ['foo', 'bar', 'is_any'])

        # uses filtr.verbs if set
        filtr.verbs = ['is_true', 'is_false']
        self.assertEqual(filtr.get_verbs(), ['is_true', 'is_false', 'is_any'])

        # may add is/null verbs
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter,
                                 nullable=True)
        self.assertEqual(filtr.get_verbs(), ['equal', 'not_equal',
                                             'is_null', 'is_not_null',
                                             'is_any'])

        # filtr.verbs can be a callable
        filtr.nullable = False
        filtr.verbs = lambda: ['baz', 'blarg']
        self.assertEqual(filtr.get_verbs(), ['baz', 'blarg', 'is_any'])

    def test_get_default_verb(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)
        self.assertFalse(hasattr(filtr, 'verbs'))
        self.assertEqual(filtr.default_verbs, ['equal', 'not_equal'])
        self.assertEqual(filtr.get_verbs(), ['equal', 'not_equal', 'is_any'])

        # returns first verb by default
        self.assertEqual(filtr.get_default_verb(), 'equal')

        # returns filtr.verb if set
        filtr.verb = 'foo'
        self.assertEqual(filtr.get_default_verb(), 'foo')

        # returns filtr.default_verb if set
        # (nb. this overrides filtr.verb since the point of this
        # method is to return the *default* verb)
        filtr.default_verb = 'bar'
        self.assertEqual(filtr.get_default_verb(), 'bar')

    def test_get_verb_labels(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)
        self.assertFalse(hasattr(filtr, 'verbs'))
        self.assertEqual(filtr.get_verbs(), ['equal', 'not_equal', 'is_any'])

        labels = filtr.get_verb_labels()
        self.assertIsInstance(labels, dict)
        self.assertEqual(labels['equal'], "equal to")
        self.assertEqual(labels['not_equal'], "not equal to")
        self.assertEqual(labels['is_any'], "is any")

    def test_get_valueless_verbs(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)
        self.assertFalse(hasattr(filtr, 'verbs'))
        self.assertEqual(filtr.get_verbs(), ['equal', 'not_equal', 'is_any'])

        verbs = filtr.get_valueless_verbs()
        self.assertIsInstance(verbs, list)
        self.assertIn('is_any', verbs)

    def test_apply_filter(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value, factory=mod.StringAlchemyFilter)

        # default verb used as fallback
        # self.assertEqual(filtr.default_verb, 'contains')
        filtr.default_verb = 'contains'
        filtr.verb = None
        with patch.object(filtr, 'filter_contains', side_effect=lambda q, v: q) as filter_contains:
            filtered_query = filtr.apply_filter(self.sample_query, value='foo')
            filter_contains.assert_called_once_with(self.sample_query, 'foo')
        self.assertIsNone(filtr.verb)

        # filter verb used as fallback
        filtr.verb = 'equal'
        with patch.object(filtr, 'filter_equal', create=True, side_effect=lambda q, v: q) as filter_equal:
            filtered_query = filtr.apply_filter(self.sample_query, value='foo')
            filter_equal.assert_called_once_with(self.sample_query, 'foo')

        # filter value used as fallback
        filtr.verb = 'contains'
        filtr.value = 'blarg'
        with patch.object(filtr, 'filter_contains', side_effect=lambda q, v: q) as filter_contains:
            filtered_query = filtr.apply_filter(self.sample_query)
            filter_contains.assert_called_once_with(self.sample_query, 'blarg')

        # error if invalid verb
        self.assertRaises(mod.VerbNotSupported, filtr.apply_filter,
                          self.sample_query, verb='doesnotexist')
        filtr.verbs = ['doesnotexist']
        self.assertRaises(mod.VerbNotSupported, filtr.apply_filter,
                          self.sample_query, verb='doesnotexist')

    def test_filter_is_any(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_any(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 9)


class TestAlchemyFilter(WebTestCase):

    def setUp(self):
        self.setup_web()

        model = self.app.model
        self.sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'ggg'},
            {'name': 'foo4', 'value': 'ggg'},
            {'name': 'foo5', 'value': 'ggg'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': None},
        ]
        for setting in self.sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        self.sample_query = self.session.query(model.Setting)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop('factory', mod.AlchemyFilter)
        kwargs['model_property'] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_filter_equal(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # not filtered for null value
        filtered_query = filtr.filter_equal(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)

        # nb. by default, *is filtered* by empty string
        filtered_query = filtr.filter_equal(self.sample_query, '')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 0)

        # filtered by value
        filtered_query = filtr.filter_equal(self.sample_query, 'ggg')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 3)

    def test_filter_not_equal(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # not filtered for empty value
        filtered_query = filtr.filter_not_equal(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)

        # nb. by default, *is filtered* by empty string
        filtered_query = filtr.filter_not_equal(self.sample_query, '')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 9)

        # filtered by value
        filtered_query = filtr.filter_not_equal(self.sample_query, 'ggg')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

    def test_filter_is_null(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_null(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 1)

    def test_filter_is_not_null(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_not_null(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 8)


class TestStringAlchemyFilter(WebTestCase):

    def setUp(self):
        self.setup_web()

        model = self.app.model
        self.sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'ggg'},
            {'name': 'foo4', 'value': 'ggg'},
            {'name': 'foo5', 'value': 'ggg'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in self.sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        self.sample_query = self.session.query(model.Setting)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop('factory', mod.StringAlchemyFilter)
        kwargs['model_property'] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_filter_contains(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # not filtered for empty value
        filtered_query = filtr.filter_contains(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        filtered_query = filtr.filter_contains(self.sample_query, '')
        self.assertIs(filtered_query, self.sample_query)

        # filtered by value
        filtered_query = filtr.filter_contains(self.sample_query, 'ggg')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 3)

    def test_filter_does_not_contain(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # not filtered for empty value
        filtered_query = filtr.filter_does_not_contain(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        filtered_query = filtr.filter_does_not_contain(self.sample_query, '')
        self.assertIs(filtered_query, self.sample_query)

        # filtered by value
        filtered_query = filtr.filter_does_not_contain(self.sample_query, 'ggg')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)


class TestBooleanAlchemyFilter(WebTestCase):

    def setUp(self):
        self.setup_web()

        model = self.app.model
        self.sample_data = [
            {'username': 'alice',
             'prevent_edit': False,
             'active': True},
            {'username': 'bob',
             'prevent_edit': True,
             'active': True},
            {'username': 'charlie',
             'active': False,
             'prevent_edit': None},
        ]
        for user in self.sample_data:
            user = model.User(**user)
            self.session.add(user)
        self.session.commit()
        self.sample_query = self.session.query(model.User)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop('factory', mod.BooleanAlchemyFilter)
        kwargs['model_property'] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_get_verbs(self):
        model = self.app.model

        # bool field, not nullable
        filtr = self.make_filter(model.User.active,
                                 factory=mod.BooleanAlchemyFilter,
                                 nullable=False)
        self.assertFalse(hasattr(filtr, 'verbs'))
        self.assertEqual(filtr.default_verbs, ['is_true', 'is_false'])

        # by default, returns default verbs (plus 'is_any')
        self.assertEqual(filtr.get_verbs(), ['is_true', 'is_false', 'is_any'])

        # default verbs can be a callable
        filtr.default_verbs = lambda: ['foo', 'bar']
        self.assertEqual(filtr.get_verbs(), ['foo', 'bar', 'is_any'])

        # bool field, *nullable*
        filtr = self.make_filter(model.User.active,
                                 factory=mod.BooleanAlchemyFilter,
                                 nullable=True)
        self.assertFalse(hasattr(filtr, 'verbs'))
        self.assertEqual(filtr.default_verbs, ['is_true', 'is_false'])

        # effective verbs also include is_false_null
        self.assertEqual(filtr.get_verbs(), ['is_true', 'is_false', 'is_false_null',
                                             'is_null', 'is_not_null', 'is_any'])

    def test_coerce_value(self):
        model = self.app.model
        filtr = self.make_filter(model.User.active)

        self.assertIsNone(filtr.coerce_value(None))

        self.assertTrue(filtr.coerce_value(True))
        self.assertTrue(filtr.coerce_value(1))
        self.assertTrue(filtr.coerce_value('1'))

        self.assertFalse(filtr.coerce_value(False))
        self.assertFalse(filtr.coerce_value(0))
        self.assertFalse(filtr.coerce_value(''))

    def test_filter_is_true(self):
        model = self.app.model
        filtr = self.make_filter(model.User.active)
        self.assertEqual(self.sample_query.count(), 3)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_true(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

    def test_filter_is_false(self):
        model = self.app.model
        filtr = self.make_filter(model.User.active)
        self.assertEqual(self.sample_query.count(), 3)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_false(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 1)

    def test_filter_is_false_null(self):
        model = self.app.model
        filtr = self.make_filter(model.User.prevent_edit)
        self.assertEqual(self.sample_query.count(), 3)

        # nb. only one account is marked with "prevent edit"
        filtered_query = filtr.filter_is_false_null(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)


class TheLocalThing(Base):
    __tablename__ = 'the_local_thing'
    id = sa.Column(sa.Integer(), primary_key=True, autoincrement=False)
    date = sa.Column(sa.DateTime(timezone=True), nullable=True)


class TestDateAlchemyFilter(WebTestCase):

    def setUp(self):
        self.setup_web()
        model = self.app.model

        # nb. create table for TheLocalThing
        model.Base.metadata.create_all(bind=self.session.bind)

        self.sample_data = [
            {'id': 1, 'date': datetime.date(2024, 1, 1)},
            {'id': 2, 'date': datetime.date(2024, 1, 1)},
            {'id': 3, 'date': datetime.date(2024, 3, 1)},
            {'id': 4, 'date': datetime.date(2024, 3, 1)},
            {'id': 5, 'date': None},
            {'id': 6, 'date': None},
        ]

        for thing in self.sample_data:
            thing = TheLocalThing(**thing)
            self.session.add(thing)
        self.session.commit()

        self.sample_query = self.session.query(TheLocalThing)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop('factory', mod.DateAlchemyFilter)
        kwargs['model_property'] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_coerce_value(self):
        filtr = self.make_filter(TheLocalThing.date)

        # null value
        self.assertIsNone(filtr.coerce_value(None))

        # value as datetime
        value = datetime.date(2024, 1, 1)
        result = filtr.coerce_value(value)
        self.assertIs(value, result)

        # value as string
        result = filtr.coerce_value('2024-04-01')
        self.assertIsInstance(result, datetime.date)
        self.assertEqual(result, datetime.date(2024, 4, 1))

        # invalid
        result = filtr.coerce_value('thisinputisbad')
        self.assertIsNone(result)

    def test_greater_than(self):
        model = self.app.model

        filtr = self.make_filter(TheLocalThing.date)
        self.assertEqual(self.sample_query.count(), 6)

        # null value ignored
        filtered_query = filtr.filter_greater_than(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

        # value as date
        filtered_query = filtr.filter_greater_than(self.sample_query, datetime.date(2024, 2, 1))
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as string
        filtered_query = filtr.filter_greater_than(self.sample_query, '2024-02-01')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

    def test_greater_equal(self):
        model = self.app.model

        filtr = self.make_filter(TheLocalThing.date)
        self.assertEqual(self.sample_query.count(), 6)

        # null value ignored
        filtered_query = filtr.filter_greater_equal(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

        # value as date (clear of boundary)
        filtered_query = filtr.filter_greater_equal(self.sample_query, datetime.date(2024, 2, 1))
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as date (at boundary)
        filtered_query = filtr.filter_greater_equal(self.sample_query, datetime.date(2024, 3, 1))
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as string
        filtered_query = filtr.filter_greater_equal(self.sample_query, '2024-01-01')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 4)

    def test_less_than(self):
        model = self.app.model

        filtr = self.make_filter(TheLocalThing.date)
        self.assertEqual(self.sample_query.count(), 6)

        # null value ignored
        filtered_query = filtr.filter_less_than(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

        # value as date
        filtered_query = filtr.filter_less_than(self.sample_query, datetime.date(2024, 2, 1))
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as string
        filtered_query = filtr.filter_less_than(self.sample_query, '2024-04-01')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 4)

    def test_less_equal(self):
        model = self.app.model

        filtr = self.make_filter(TheLocalThing.date)
        self.assertEqual(self.sample_query.count(), 6)

        # null value ignored
        filtered_query = filtr.filter_less_equal(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

        # value as date (clear of boundary)
        filtered_query = filtr.filter_less_equal(self.sample_query, datetime.date(2024, 2, 1))
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as date (at boundary)
        filtered_query = filtr.filter_less_equal(self.sample_query, datetime.date(2024, 3, 1))
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as string
        filtered_query = filtr.filter_less_equal(self.sample_query, '2024-04-01')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 4)


class TestVerbNotSupported(TestCase):

    def test_basic(self):
        error = mod.VerbNotSupported('equal')
        self.assertEqual(str(error), "unknown filter verb not supported: equal")



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Common Views
"""

import logging

import colander
from pyramid.renderers import render

from wuttaweb.views import View
from wuttaweb.forms import widgets
from wuttaweb.db import Session


log = logging.getLogger(__name__)


class CommonView(View):
    """
    Common views shared by all apps.
    """

    def home(self, session=None):
        """
        Home page view.

        Template: ``/home.mako``

        This is normally the view shown when a user navigates to the
        root URL for the web app.
        """
        model = self.app.model
        session = session or Session()

        # nb. redirect to /setup if no users exist
        user = session.query(model.User).first()
        if not user:
            return self.redirect(self.request.route_url('setup'))

        # maybe auto-redirect anons to login
        if not self.request.user:
            if self.config.get_bool('wuttaweb.home_redirect_to_login'):
                return self.redirect(self.request.route_url('login'))

        return {
            'index_title': self.app.get_title(),
        }

    def forbidden_view(self):
        """
        This view is shown when a request triggers a 403 Forbidden error.

        Template: ``/forbidden.mako``
        """
        return {'index_title': self.app.get_title()}

    def notfound_view(self):
        """
        This view is shown when a request triggers a 404 Not Found error.

        Template: ``/notfound.mako``
        """
        return {'index_title': self.app.get_title()}

    def feedback(self):
        """ """
        model = self.app.model
        session = Session()

        # validate form
        schema = self.feedback_make_schema()
        form = self.make_form(schema=schema)
        if not form.validate():
            # TODO: native Form class should better expose error(s)
            dform = form.get_deform()
            return {'error': str(dform.error)}

        # build email template context
        context = dict(form.validated)
        if context['user_uuid']:
            context['user'] = session.get(model.User, context['user_uuid'])
            context['user_url'] = self.request.route_url('users.view', uuid=context['user_uuid'])
        context['client_ip'] = self.request.client_addr

        # send email
        try:
            self.feedback_send(context)
        except Exception as error:
            log.warning("failed to send feedback email", exc_info=True)
            return {'error': str(error) or error.__class__.__name__}

        return {'ok': True}

    def feedback_make_schema(self):
        """ """
        schema = colander.Schema()

        schema.add(colander.SchemaNode(colander.String(),
                                       name='referrer'))

        schema.add(colander.SchemaNode(colander.String(),
                                       name='user_uuid',
                                       missing=None))

        schema.add(colander.SchemaNode(colander.String(),
                                       name='user_name'))

        schema.add(colander.SchemaNode(colander.String(),
                                       name='message'))

        return schema

    def feedback_send(self, context):
        """ """
        self.app.send_email('feedback', context)

    def setup(self, session=None):
        """
        View for first-time app setup, to create admin user.

        Template: ``/setup.mako``

        This page is only meant for one-time use.  As such, if the app
        DB contains any users, this page will always redirect to the
        home page.

        However if no users exist yet, this will show a form which may
        be used to create the first admin user.  When finished, user
        will be redirected to the login page.

        .. note::

           As long as there are no users in the DB, both the home and
           login pages will automatically redirect to this one.
        """
        model = self.app.model
        session = session or Session()

        # nb. this view only available until first user is created
        user = session.query(model.User).first()
        if user:
            return self.redirect(self.request.route_url('home'))

        form = self.make_form(fields=['username', 'password', 'first_name', 'last_name'],
                              show_button_cancel=False,
                              show_button_reset=True)
        form.set_widget('password', widgets.CheckedPasswordWidget())
        form.set_required('first_name', False)
        form.set_required('last_name', False)

        if form.validate():
            auth = self.app.get_auth_handler()
            data = form.validated

            # make user
            user = auth.make_user(session=session, username=data['username'])
            auth.set_user_password(user, data['password'])

            # assign admin role
            admin = auth.get_role_administrator(session)
            user.roles.append(admin)
            admin.notes = ("users in this role may \"become root\".\n\n"
                           "it's recommended not to grant other perms to this role.")

            # initialize built-in roles
            authed = auth.get_role_authenticated(session)
            authed.notes = ("this role represents any user who *is* logged in.\n\n"
                            "you may grant any perms you like to it.")
            anon = auth.get_role_anonymous(session)
            anon.notes = ("this role represents any user who is *not* logged in.\n\n"
                          "you may grant any perms you like to it.")

            # also make "Site Admin" role
            site_admin_perms = [
                'appinfo.list',
                'appinfo.configure',
                'people.list',
                'people.create',
                'people.view',
                'people.edit',
                'people.delete',
                'roles.list',
                'roles.create',
                'roles.view',
                'roles.edit',
                'roles.edit_builtin',
                'roles.delete',
                'settings.list',
                'settings.create',
                'settings.view',
                'settings.edit',
                'settings.delete',
                'settings.delete_bulk',
                'upgrades.list',
                'upgrades.create',
                'upgrades.view',
                'upgrades.edit',
                'upgrades.delete',
                'upgrades.execute',
                'upgrades.download',
                'upgrades.configure',
                'users.list',
                'users.create',
                'users.view',
                'users.edit',
                'users.delete',
            ]
            admin2 = model.Role(name="Site Admin")
            admin2.notes = ("this is the \"daily driver\" admin role.\n\n"
                            "you may grant any perms you like to it.")
            session.add(admin2)
            user.roles.append(admin2)
            for perm in site_admin_perms:
                auth.grant_permission(admin2, perm)

            # maybe make person
            if data['first_name'] or data['last_name']:
                first = data['first_name']
                last = data['last_name']
                person = model.Person(first_name=first,
                                      last_name=last,
                                      full_name=(f"{first} {last}").strip())
                session.add(person)
                user.person = person

            # send user to /login
            self.request.session.flash("Account created! Please login below.")
            return self.redirect(self.request.route_url('login'))

        return {
            'index_title': self.app.get_title(),
            'form': form,
        }

    @classmethod
    def defaults(cls, config):
        cls._defaults(config)

    @classmethod
    def _defaults(cls, config):

        config.add_wutta_permission_group('common', "(common)", overwrite=False)

        # home page
        config.add_route('home', '/')
        config.add_view(cls, attr='home',
                        route_name='home',
                        renderer='/home.mako')

        # forbidden
        config.add_forbidden_view(cls, attr='forbidden_view',
                                  renderer='/forbidden.mako')

        # notfound
        # nb. also, auto-correct URLs which require trailing slash
        config.add_notfound_view(cls, attr='notfound_view',
                                 append_slash=True,
                                 renderer='/notfound.mako')

        # feedback
        config.add_route('feedback', '/feedback',
                         request_method='POST')
        config.add_view(cls, attr='feedback',
                        route_name='feedback',
                        permission='common.feedback',
                        renderer='json')
        config.add_wutta_permission('common', 'common.feedback',
                                    "Send user feedback about the app")

        # setup
        config.add_route('setup', '/setup')
        config.add_view(cls, attr='setup',
                        route_name='setup',
                        renderer='/setup.mako')


def defaults(config, **kwargs):
    base = globals()

    CommonView = kwargs.get('CommonView', base['CommonView'])
    CommonView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-

from unittest import TestCase
from unittest.mock import MagicMock

import fanstatic
from pyramid import testing

from wuttjamaican.conf import WuttaConfig
from wuttjamaican.testing import FileConfigTestCase
from wuttaweb import subscribers
from wuttaweb.menus import MenuHandler


class DataTestCase(FileConfigTestCase):
    """
    Base class for test suites requiring a full (typical) database.
    """

    def setUp(self):
        self.setup_db()

    def setup_db(self):
        self.setup_files()
        self.config = WuttaConfig(defaults={
            'wutta.db.default.url': 'sqlite://',
        })
        self.app = self.config.get_app()

        # init db
        model = self.app.model
        model.Base.metadata.create_all(bind=self.config.appdb_engine)
        self.session = self.app.make_session()

    def tearDown(self):
        self.teardown_db()

    def teardown_db(self):
        self.teardown_files()


class WebTestCase(DataTestCase):
    """
    Base class for test suites requiring a full (typical) web app.
    """

    def setUp(self):
        self.setup_web()

    def setup_web(self):
        self.setup_db()
        self.request = self.make_request()
        self.pyramid_config = testing.setUp(request=self.request, settings={
            'wutta_config': self.config,
            'mako.directories': ['wuttaweb:templates'],
            'pyramid_deform.template_search_path': 'wuttaweb:templates/deform',
        })

        # init web
        self.pyramid_config.include('pyramid_deform')
        self.pyramid_config.include('pyramid_mako')
        self.pyramid_config.add_directive('add_wutta_permission_group',
                                          'wuttaweb.auth.add_permission_group')
        self.pyramid_config.add_directive('add_wutta_permission',
                                          'wuttaweb.auth.add_permission')
        self.pyramid_config.add_subscriber('wuttaweb.subscribers.before_render',
                                           'pyramid.events.BeforeRender')
        self.pyramid_config.include('wuttaweb.static')

        # nb. mock out fanstatic env..good enough for now to avoid errors..
        needed = fanstatic.init_needed()
        self.request.environ[fanstatic.NEEDED] = needed

        # setup new request w/ anonymous user
        event = MagicMock(request=self.request)
        subscribers.new_request(event)
        def user_getter(request, **kwargs): pass
        subscribers.new_request_set_user(event, db_session=self.session,
                                         user_getter=user_getter)

    def tearDown(self):
        self.teardown_web()

    def teardown_web(self):
        testing.tearDown()
        self.teardown_db()

    def make_request(self):
        return testing.DummyRequest()


class NullMenuHandler(MenuHandler):
    """
    Dummy menu handler for testing.
    """
    def make_menus(self, request, **kwargs):
        return []



# -*- coding: utf-8; -*-

from wuttaweb.views import essential as mod
from tests.util import WebTestCase


class TestEssentialViews(WebTestCase):

    def test_includeme(self):
        self.pyramid_config.include('wuttaweb.views.essential')



# -*- coding: utf-8; -*-

from unittest.mock import patch

import colander

from wuttaweb.views import common as mod
from tests.util import WebTestCase


class TestCommonView(WebTestCase):

    def make_view(self):
        return mod.CommonView(self.request)

    def test_includeme(self):
        self.pyramid_config.include('wuttaweb.views.common')

    def test_forbidden_view(self):
        view = self.make_view()
        context = view.forbidden_view()
        self.assertEqual(context['index_title'], self.app.get_title())

    def test_notfound_view(self):
        view = self.make_view()
        context = view.notfound_view()
        self.assertEqual(context['index_title'], self.app.get_title())

    def test_home(self):
        self.pyramid_config.add_route('setup', '/setup')
        self.pyramid_config.add_route('login', '/login')
        model = self.app.model
        view = self.make_view()

        # if no users then home page will redirect
        response = view.home(session=self.session)
        self.assertEqual(response.status_code, 302)

        # so add a user
        user = model.User(username='foo')
        self.session.add(user)
        self.session.commit()

        # now we see the home page
        context = view.home(session=self.session)
        self.assertEqual(context['index_title'], self.app.get_title())

        # but if configured, anons will be redirected to login
        self.config.setdefault('wuttaweb.home_redirect_to_login', 'true')
        response = view.home(session=self.session)
        self.assertEqual(response.status_code, 302)

        # now only an auth'ed user can see home page
        self.request.user = user
        context = view.home(session=self.session)
        self.assertEqual(context['index_title'], self.app.get_title())

    def test_feedback_make_schema(self):
        view = self.make_view()
        schema = view.feedback_make_schema()
        self.assertIsInstance(schema, colander.Schema)
        self.assertIn('message', schema)

    def test_feedback(self):
        self.pyramid_config.add_route('users.view', '/users/{uuid}')
        model = self.app.model
        user = model.User(username='barney')
        self.session.add(user)
        self.session.commit()

        view = self.make_view()
        with patch.object(view, 'feedback_send') as feedback_send:

            # basic send, no user
            self.request.client_addr = '127.0.0.1'
            self.request.method = 'POST'
            self.request.POST = {
                'referrer': '/foo',
                'user_name': "Barney Rubble",
                'message': "hello world",
            }
            context = view.feedback()
            self.assertEqual(context, {'ok': True})
            feedback_send.assert_called_once()

            # reset
            feedback_send.reset_mock()

            # basic send, with user
            self.request.user = user
            self.request.POST['user_uuid'] = str(user.uuid)
            with patch.object(mod, 'Session', return_value=self.session):
                context = view.feedback()
            self.assertEqual(context, {'ok': True})
            feedback_send.assert_called_once()

            # reset
            self.request.user = None
            feedback_send.reset_mock()

            # invalid form data
            self.request.POST = {'message': 'hello world'}
            context = view.feedback()
            self.assertEqual(list(context), ['error'])
            self.assertIn('Required', context['error'])
            feedback_send.assert_not_called()

            # error on send
            self.request.POST = {
                'referrer': '/foo',
                'user_name': "Barney Rubble",
                'message': "hello world",
            }
            feedback_send.side_effect = RuntimeError
            context = view.feedback()
            feedback_send.assert_called_once()
            self.assertEqual(list(context), ['error'])
            self.assertIn('RuntimeError', context['error'])

    def test_feedback_send(self):
        view = self.make_view()
        with patch.object(self.app, 'send_email') as send_email:
            view.feedback_send({'user_name': "Barney",
                                'message': "hello world"})
            send_email.assert_called_once_with('feedback', {
                'user_name': "Barney",
                'message': "hello world"
            })

    def test_setup(self):
        self.pyramid_config.add_route('home', '/')
        self.pyramid_config.add_route('login', '/login')
        model = self.app.model
        auth = self.app.get_auth_handler()
        view = self.make_view()

        # at first, can see the setup page
        self.assertEqual(self.session.query(model.User).count(), 0)
        context = view.setup(session=self.session)
        self.assertEqual(context['index_title'], self.app.get_title())

        # so add a user
        user = model.User(username='foo')
        self.session.add(user)
        self.session.commit()

        # once user exists it will always redirect
        response = view.setup(session=self.session)
        self.assertEqual(response.status_code, 302)

        # delete that user
        self.session.delete(user)
        self.session.commit()

        # so we can see the setup page again
        context = view.setup(session=self.session)
        self.assertEqual(context['index_title'], self.app.get_title())

        # and finally, post data to create admin user
        self.request.method = 'POST'
        self.request.POST = {
            'username': 'barney',
            '__start__': 'password:mapping',
            'password': 'testpass',
            'password-confirm': 'testpass',
            '__end__': 'password:mapping',
            'first_name': "Barney",
            'last_name': "Rubble",
        }
        response = view.setup(session=self.session)
        # nb. redirects on success
        self.assertEqual(response.status_code, 302)
        barney = self.session.query(model.User).one()
        self.assertEqual(barney.username, 'barney')
        self.assertTrue(auth.check_user_password(barney, 'testpass'))
        admin = auth.get_role_administrator(self.session)
        self.assertIn(admin, barney.roles)
        self.assertIsNotNone(barney.person)
        person = barney.person
        self.assertEqual(person.first_name, "Barney")
        self.assertEqual(person.last_name, "Rubble")
        self.assertEqual(person.full_name, "Barney Rubble")



# -*- coding: utf-8; -*-

from pyramid.httpexceptions import HTTPFound, HTTPForbidden, HTTPNotFound

from wuttaweb.views import base as mod
from wuttaweb.forms import Form
from wuttaweb.grids import Grid, GridAction
from tests.util import WebTestCase


class TestView(WebTestCase):

    def make_view(self):
        return mod.View(self.request)

    def test_basic(self):
        view = self.make_view()
        self.assertIs(view.request, self.request)
        self.assertIs(view.config, self.config)
        self.assertIs(view.app, self.app)

    def test_forbidden(self):
        view = self.make_view()
        error = view.forbidden()
        self.assertIsInstance(error, HTTPForbidden)

    def test_make_form(self):
        view = self.make_view()
        form = view.make_form()
        self.assertIsInstance(form, Form)

    def test_make_grid(self):
        view = self.make_view()
        grid = view.make_grid()
        self.assertIsInstance(grid, Grid)

    def test_make_grid_action(self):
        view = self.make_view()
        action = view.make_grid_action('view')
        self.assertIsInstance(action, GridAction)

    def test_notfound(self):
        view = self.make_view()
        error = view.notfound()
        self.assertIsInstance(error, HTTPNotFound)

    def test_redirect(self):
        view = self.make_view()
        error = view.redirect('/')
        self.assertIsInstance(error, HTTPFound)
        self.assertEqual(error.location, '/')

    def test_file_response(self):
        view = self.make_view()

        # default uses attachment behavior
        datfile = self.write_file('dat.txt', 'hello')
        response = view.file_response(datfile)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content_disposition, 'attachment; filename="dat.txt"')

        # but can disable attachment behavior
        datfile = self.write_file('dat.txt', 'hello')
        response = view.file_response(datfile, attachment=False)
        self.assertEqual(response.status_code, 200)
        self.assertIsNone(response.content_disposition)

        # path not found
        crapfile = '/does/not/exist'
        response = view.file_response(crapfile)
        self.assertEqual(response.status_code, 404)

    def test_json_response(self):
        view = self.make_view()
        response = view.json_response({'foo': 'bar'})
        self.assertEqual(response.status_code, 200)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database sessions for web app

The web app uses a different database session than other
(e.g. console) apps.  The web session is "registered" to the HTTP
request/response life cycle (aka. transaction) such that the session
is automatically rolled back on error, and automatically committed if
the response is finalized without error.

.. class:: Session

   Primary database session class for the web app.

   Note that you often do not need to "instantiate" this session, and
   can instead call methods directly on the class::

      from wuttaweb.db import Session

      users = Session.query(model.User).all()

   However in certain cases you may still want/need to instantiate it,
   e.g. when passing a "true/normal" session to other logic.  But you
   can always call instance methods as well::

      from wuttaweb.db import Session
      from some_place import some_func

      session = Session()

      # nb. assuming func does not expect a "web" session per se, pass instance
      some_func(session)

      # nb. these behave the same (instance vs. class method)
      users = session.query(model.User).all()
      users = Session.query(model.User).all()
"""

from sqlalchemy import orm
from zope.sqlalchemy.datamanager import register


Session = orm.scoped_session(orm.sessionmaker())

register(Session)



# -*- coding: utf-8; -*-

from unittest.mock import patch

from sqlalchemy import orm

from pyramid.httpexceptions import HTTPNotFound

from wuttaweb.views import people
from tests.util import WebTestCase


class TestPersonView(WebTestCase):

    def make_view(self):
        return people.PersonView(self.request)

    def test_includeme(self):
        self.pyramid_config.include('wuttaweb.views.people')

    def test_get_query(self):
        view = self.make_view()
        query = view.get_query(session=self.session)
        self.assertIsInstance(query, orm.Query)

    def test_configure_grid(self):
        model = self.app.model
        view = self.make_view()
        grid = view.make_grid(model_class=model.Setting)
        self.assertEqual(grid.linked_columns, [])
        view.configure_grid(grid)
        self.assertIn('full_name', grid.linked_columns)

    def test_configure_form(self):
        model = self.app.model
        view = self.make_view()
        form = view.make_form(model_class=model.Person)

        # required fields
        with patch.object(view, 'creating', new=True):
            form.set_fields(form.get_model_fields())
            self.assertEqual(form.required_fields, {})
            view.configure_form(form)
            self.assertTrue(form.required_fields)
            self.assertFalse(form.required_fields['middle_name'])

        person = model.Person(full_name="Barney Rubble")
        user = model.User(username='barney', person=person)
        self.session.add(user)
        self.session.commit()

        # users field
        with patch.object(view, 'viewing', new=True):
            form = view.make_form(model_instance=person)
            self.assertEqual(form.defaults, {})
            view.configure_form(form)
            self.assertIn('_users', form.defaults)

    def test_autocomplete_query(self):
        model = self.app.model

        person1 = model.Person(full_name="George Jones")
        self.session.add(person1)
        person2 = model.Person(full_name="George Strait")
        self.session.add(person2)
        self.session.commit()

        view = self.make_view()
        with patch.object(view, 'Session', return_value=self.session):

            # both people match
            query = view.autocomplete_query('george')
            self.assertEqual(query.count(), 2)

            # just 1 match
            query = view.autocomplete_query('jones')
            self.assertEqual(query.count(), 1)

            # no matches
            query = view.autocomplete_query('sally')
            self.assertEqual(query.count(), 0)

    def test_view_profile(self):
        self.pyramid_config.include('wuttaweb.views.common')
        self.pyramid_config.include('wuttaweb.views.auth')
        self.pyramid_config.add_route('people', '/people/')

        model = self.app.model
        person = model.Person(full_name="Barney Rubble")
        self.session.add(person)
        self.session.commit()

        # sanity check
        view = self.make_view()
        self.request.matchdict = {'uuid': person.uuid}
        response = view.view_profile(session=self.session)
        self.assertEqual(response.status_code, 200)

    def test_make_user(self):
        self.pyramid_config.include('wuttaweb.views.common')

        model = self.app.model
        person = model.Person(full_name="Barney Rubble")
        self.session.add(person)
        self.session.commit()

        # sanity check
        view = self.make_view()
        self.request.matchdict = {'uuid': person.uuid}
        response = view.make_user()
        # nb. this always redirects for now
        self.assertEqual(response.status_code, 302)



# -*- coding: utf-8; -*-

import json
from unittest import TestCase
from unittest.mock import MagicMock, patch

from wuttjamaican.conf import WuttaConfig

from pyramid import testing
from pyramid.security import remember

from wuttaweb import subscribers
from wuttaweb import helpers
from wuttaweb.auth import WuttaSecurityPolicy


class TestNewRequest(TestCase):

    def setUp(self):
        self.config = WuttaConfig()
        self.request = self.make_request()
        self.pyramid_config = testing.setUp(request=self.request, settings={
            'wutta_config': self.config,
        })

    def tearDown(self):
        testing.tearDown()

    def make_request(self):
        request = testing.DummyRequest()
        # request.registry.settings = {'wutta_config': self.config}
        return request

    def test_wutta_config(self):
        event = MagicMock(request=self.request)

        # request gets a new attr
        self.assertFalse(hasattr(self.request, 'wutta_config'))
        subscribers.new_request(event)
        self.assertTrue(hasattr(self.request, 'wutta_config'))
        self.assertIs(self.request.wutta_config, self.config)

    def test_use_oruga_default(self):
        event = MagicMock(request=self.request)

        # request gets a new attr, false by default
        self.assertFalse(hasattr(self.request, 'use_oruga'))
        subscribers.new_request(event)
        self.assertFalse(self.request.use_oruga)

    def test_use_oruga_custom(self):
        self.config.setdefault('wuttaweb.oruga_detector.spec', 'tests.test_subscribers:custom_oruga_detector')
        event = MagicMock(request=self.request)

        # request gets a new attr, which should be true
        self.assertFalse(hasattr(self.request, 'use_oruga'))
        subscribers.new_request(event)
        self.assertTrue(self.request.use_oruga)

    def test_get_referrer(self):
        event = MagicMock(request=self.request)

        def home(request):
            pass

        self.pyramid_config.add_route('home', '/')
        self.pyramid_config.add_view(home, route_name='home')

        self.assertFalse(hasattr(self.request, 'get_referrer'))
        subscribers.new_request(event)
        self.assertTrue(hasattr(self.request, 'get_referrer'))

        # default if no referrer, is home route
        url = self.request.get_referrer()
        self.assertEqual(url, self.request.route_url('home'))

        # can specify another default
        url = self.request.get_referrer(default='https://wuttaproject.org')
        self.assertEqual(url, 'https://wuttaproject.org')

        # or referrer can come from user session
        self.request.session['referrer'] = 'https://rattailproject.org'
        self.assertIn('referrer', self.request.session)
        url = self.request.get_referrer()
        self.assertEqual(url, 'https://rattailproject.org')
        # nb. referrer should also have been removed from user session
        self.assertNotIn('referrer', self.request.session)

        # or referrer can come from request params
        self.request.params['referrer'] = 'https://kernel.org'
        url = self.request.get_referrer()
        self.assertEqual(url, 'https://kernel.org')


def custom_oruga_detector(request):
    return True


class TestNewRequestSetUser(TestCase):

    def setUp(self):
        self.config = WuttaConfig(defaults={
            'wutta.db.default.url': 'sqlite://',
        })

        self.request = testing.DummyRequest(wutta_config=self.config)
        self.pyramid_config = testing.setUp(request=self.request, settings={
            'wutta_config': self.config,
        })

        self.app = self.config.get_app()
        model = self.app.model
        model.Base.metadata.create_all(bind=self.config.appdb_engine)
        self.session = self.app.make_session()
        self.user = model.User(username='barney')
        self.session.add(self.user)
        self.session.commit()

        self.pyramid_config.set_security_policy(WuttaSecurityPolicy(db_session=self.session))

    def tearDown(self):
        testing.tearDown()

    def test_anonymous(self):
        self.assertFalse(hasattr(self.request, 'user'))
        event = MagicMock(request=self.request)
        subscribers.new_request_set_user(event)
        self.assertIsNone(self.request.user)

    def test_authenticated(self):
        uuid = self.user.uuid
        self.assertIsNotNone(uuid)
        remember(self.request, uuid)
        event = MagicMock(request=self.request)
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertIs(self.request.user, self.user)

    def test_is_admin(self):
        event = MagicMock(request=self.request)

        # anonymous user
        self.assertFalse(hasattr(self.request, 'user'))
        self.assertFalse(hasattr(self.request, 'is_admin'))
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertIsNone(self.request.user)
        self.assertFalse(self.request.is_admin)

        # reset
        del self.request.is_admin

        # authenticated user, but still not an admin
        self.request.user = self.user
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertIs(self.request.user, self.user)
        self.assertFalse(self.request.is_admin)

        # reset
        del self.request.is_admin

        # but if we make them an admin, it changes
        auth = self.app.get_auth_handler()
        admin = auth.get_role_administrator(self.session)
        self.user.roles.append(admin)
        self.session.commit()
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertIs(self.request.user, self.user)
        self.assertTrue(self.request.is_admin)

    def test_is_root(self):
        event = MagicMock(request=self.request)

        # anonymous user
        self.assertFalse(hasattr(self.request, 'user'))
        self.assertFalse(hasattr(self.request, 'is_root'))
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertIsNone(self.request.user)
        self.assertFalse(self.request.is_root)

        # reset
        del self.request.is_admin
        del self.request.is_root

        # authenticated user, but still not an admin
        self.request.user = self.user
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertIs(self.request.user, self.user)
        self.assertFalse(self.request.is_root)

        # reset
        del self.request.is_admin
        del self.request.is_root

        # even if we make them an admin, still not yet root
        auth = self.app.get_auth_handler()
        admin = auth.get_role_administrator(self.session)
        self.user.roles.append(admin)
        self.session.commit()
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertIs(self.request.user, self.user)
        self.assertTrue(self.request.is_admin)
        self.assertFalse(self.request.is_root)

        # reset
        del self.request.is_admin
        del self.request.is_root

        # root status flag lives in user session
        self.request.session['is_root'] = True
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertTrue(self.request.is_admin)
        self.assertTrue(self.request.is_root)

    def test_user_permissions(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        event = MagicMock(request=self.request)

        # anonymous user
        self.assertFalse(hasattr(self.request, 'user_permissions'))
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertEqual(self.request.user_permissions, set())

        # reset
        del self.request.user_permissions

        # add user to role with perms
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        auth.grant_permission(blokes, 'appinfo.list')
        self.user.roles.append(blokes)
        self.session.commit()

        # authenticated user, with perms
        self.request.user = self.user
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertEqual(self.request.user_permissions, {'appinfo.list'})

    def test_has_perm(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        event = MagicMock(request=self.request)

        # anonymous user
        self.assertFalse(hasattr(self.request, 'has_perm'))
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertFalse(self.request.has_perm('appinfo.list'))

        # reset
        del self.request.user_permissions
        del self.request.has_perm
        del self.request.has_any_perm

        # add user to role with perms
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        auth.grant_permission(blokes, 'appinfo.list')
        self.user.roles.append(blokes)
        self.session.commit()

        # authenticated user, with perms
        self.request.user = self.user
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertTrue(self.request.has_perm('appinfo.list'))

        # reset
        del self.request.user_permissions
        del self.request.has_perm
        del self.request.has_any_perm

        # drop user from role, no more perms
        self.user.roles.remove(blokes)
        self.session.commit()
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertFalse(self.request.has_perm('appinfo.list'))

        # reset
        del self.request.user_permissions
        del self.request.has_perm
        del self.request.has_any_perm
        del self.request.is_admin
        del self.request.is_root

        # root user always has perms
        admin = auth.get_role_administrator(self.session)
        self.user.roles.append(admin)
        self.session.commit()
        self.request.session['is_root'] = True
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertTrue(self.request.has_perm('appinfo.list'))

    def test_has_any_perm(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        event = MagicMock(request=self.request)

        # anonymous user
        self.assertFalse(hasattr(self.request, 'has_any_perm'))
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertFalse(self.request.has_any_perm('appinfo.list'))

        # reset
        del self.request.user_permissions
        del self.request.has_perm
        del self.request.has_any_perm

        # add user to role with perms
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        auth.grant_permission(blokes, 'appinfo.list')
        self.user.roles.append(blokes)
        self.session.commit()

        # authenticated user, with perms
        self.request.user = self.user
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertTrue(self.request.has_any_perm('appinfo.list', 'appinfo.view'))

        # reset
        del self.request.user_permissions
        del self.request.has_perm
        del self.request.has_any_perm

        # drop user from role, no more perms
        self.user.roles.remove(blokes)
        self.session.commit()
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertFalse(self.request.has_any_perm('appinfo.list'))

        # reset
        del self.request.user_permissions
        del self.request.has_perm
        del self.request.has_any_perm
        del self.request.is_admin
        del self.request.is_root

        # root user always has perms
        admin = auth.get_role_administrator(self.session)
        self.user.roles.append(admin)
        self.session.commit()
        self.request.session['is_root'] = True
        subscribers.new_request_set_user(event, db_session=self.session)
        self.assertTrue(self.request.has_any_perm('appinfo.list'))


class TestBeforeRender(TestCase):

    def setUp(self):
        self.config = WuttaConfig(defaults={
            'wutta.web.menus.handler_spec': 'tests.util:NullMenuHandler',
        })

    def make_request(self):
        request = testing.DummyRequest(use_oruga=False)
        request.registry.settings = {'wutta_config': self.config}
        request.wutta_config = self.config
        return request

    def test_basic(self):
        request = self.make_request()
        event = {'request': request}

        # event dict will get populated with more context
        subscribers.before_render(event)

        self.assertIn('config', event)
        self.assertIs(event['config'], self.config)

        self.assertIn('app', event)
        self.assertIs(event['app'], self.config.get_app())

        self.assertIn('h', event)
        self.assertIs(event['h'], helpers)

        self.assertIn('url', event)
        # TODO: not sure how to test this?
        # self.assertIs(event['url'], request.route_url)

        self.assertIn('json', event)
        self.assertIs(event['json'], json)


class TestIncludeMe(TestCase):

    def test_basic(self):
        with testing.testConfig() as pyramid_config:

            # just ensure no error happens when included..
            pyramid_config.include('wuttaweb.subscribers')



# -*- coding: utf-8; -*-

from tests.util import WebTestCase


class TestIncludeMe(WebTestCase):

    def test_basic(self):
        # just ensure no error happens when included..
        self.pyramid_config.include('wuttaweb.views')
